import os
import setuptools
from Cython.Build import cythonize
from Cython.Distutils import build_ext
import Cython.Compiler.Options

Cython.Compiler.Options.annotate = True
import numpy as np

ext_modules = [
    setuptools.Extension(
        "tt_sketch.drm.fast_lazy_gaussian",
        [os.path.join("tt_sketch", "drm", "fast_lazy_gaussian.pyx")],
        extra_compile_args=["-fopenmp"],
        extra_link_args=["-fopenmp"],
        include_dirs=[np.get_include(),"tt_sketch/drm"],
    )
]

short_description = """
Streaming randomized TT-decompositions for structured tensors.
"""


long_description = short_description

setuptools.setup(
    name="tt_sketch",
    version="1.1",
    author="Rik Voorhaar",
    description=short_description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/RikVoorhaar/tt-sketch",
    cmdclass={"build_ext": build_ext},
    ext_modules=cythonize(ext_modules),
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={"": ["*.pyx"]},
    install_requires=[
        "numpy",
        "scipy",
    ],
    setup_requires=["pytest-runner", "cython","numpy","scipy"],
    tests_require=[
        "mypy",
        "pycodestyle",
        "pytest",
        "pytest-cov",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="mathematics machine-learning tensors linear-algebra",
)



import sys

import numpy as np
import pytest

# sys.path.append("..")

import itertools
from tt_sketch.drm import (
    ALL_DRM,
    SparseGaussianDRM,
    TensorTrainDRM,
    DenseGaussianDRM,
)
from tt_sketch.drm_base import CanSlice, CanIncreaseRank
from tt_sketch.sketching_methods.abstract_methods import (
    CansketchSparse,
    CansketchTT,
    CansketchCP,
)
from tt_sketch.tensor import (
    SparseTensor,
    TensorTrain,
    CPTensor,
    TensorSum,
    DenseTensor,
    TuckerTensor,
)
from tt_sketch.sketch_dispatch import SketchMethod
from tt_sketch.sketch import (
    stream_sketch,
    blocked_stream_sketch,
    orthogonal_sketch,
    hmt_sketch,
)

DRM_DICT = {drm_type.__name__: drm_type for drm_type in ALL_DRM}
AVAILABLE_SKETCHING_METHODS = list(SketchMethod.__members__.values())


def general_rank_increase(
    X_dense,
    X_tensor,
    left_rank,
    right_rank,
    seed,
    left_drm_type,
    right_drm_type,
):
    shape = X_tensor.shape
    d = len(shape)

    left_drm = left_drm_type(
        left_rank, transpose=False, shape=X_dense.shape, seed=seed
    )
    right_seed = np.mod(seed + hash(str(d)), 2**32)
    right_drm = right_drm_type(
        right_rank, transpose=True, shape=X_dense.shape, seed=right_seed
    )
    tt_sketch1 = stream_sketch(
        X_tensor,
        left_rank,
        right_rank,
        left_drm=left_drm,
        right_drm=right_drm,
        seed=seed,
    )

    left_rank_p1 = (1,) + left_rank
    right_rank_p1 = right_rank + (1,)

    new_left_rank = tuple(r + 2 for r in left_rank)
    new_right_rank = tuple(r + 3 for r in right_rank)

    tt_sketch3 = tt_sketch1.increase_rank(
        X_tensor, new_left_rank, new_right_rank
    )

    new_left_rank_p1 = (1,) + new_left_rank
    new_right_rank_p1 = new_right_rank + (1,)

    left_drm2 = left_drm.increase_rank(new_left_rank)
    right_drm2 = right_drm.increase_rank(new_right_rank)

    tt_sketch2 = stream_sketch(
        X_tensor,
        new_left_rank,
        new_right_rank,
        left_drm=left_drm2,
        right_drm=right_drm2,
        seed=seed,
    )

    for tt_sketch_other in [tt_sketch2, tt_sketch3]:

        for i, (Y1, Y2) in enumerate(
            zip(tt_sketch1.Psi_cores, tt_sketch_other.Psi_cores)
        ):
            assert np.allclose(Y1, Y2[: left_rank_p1[i], :, : right_rank_p1[i]])

        for i, Y in enumerate(tt_sketch_other.Psi_cores):
            assert Y.shape == (
                new_left_rank_p1[i],
                shape[i],
                new_right_rank_p1[i],
            )
        for i, Z in enumerate(tt_sketch_other.Omega_mats):
            assert Z.shape == (new_left_rank[i], new_right_rank[i])

    # Check slicing cancels out rank increase
    left_drm4 = left_drm2.slice(None, left_rank)
    right_drm4 = right_drm2.slice(None, right_rank)

    tt_sketch4 = stream_sketch(
        X_tensor,
        left_rank,
        right_rank,
        left_drm=left_drm4,
        right_drm=right_drm4,
        seed=seed,
    )
    for i, (Y1, Y2) in enumerate(
        zip(tt_sketch1.Psi_cores, tt_sketch4.Psi_cores)
    ):
        assert np.all(Y1 == Y2)

    for i, (Y1, Y2) in enumerate(
        zip(tt_sketch1.Omega_mats, tt_sketch4.Omega_mats)
    ):
        assert np.all(Y1 == Y2)


# TODO: Make particular tests for slicing, right now general_blocked_sketch does
# test slicing, but only indirectly


def general_blocked_sketch(X_tensor, seed, left_drm_type, right_drm_type):
    shape = X_tensor.shape
    n_ranks = len(shape) - 1

    left_rank = (9,) * n_ranks
    right_rank = (8,) * n_ranks
    sketched_tt, left_drm, right_drm = stream_sketch(
        X_tensor,
        left_rank,
        right_rank,
        seed=seed,
        left_drm_type=left_drm_type,
        right_drm_type=right_drm_type,
        return_drm=True,
    )

    slice_left0 = (0,) * n_ranks
    slice_left1 = (3,) * n_ranks
    slice_left2 = (6,) * n_ranks
    slice_left3 = (9,) * n_ranks

    slice_right0 = (0,) * n_ranks
    slice_right1 = (4,) * n_ranks
    slice_right2 = (6,) * n_ranks
    slice_right3 = (8,) * n_ranks

    left_slices = [
        (slice_left0, slice_left1, slice_left2, slice_left3),
        (slice_left0, slice_left1, slice_left3),
        (slice_left0, slice_left3),
    ]

    right_slices = [
        (slice_right0, slice_right1, slice_right2, slice_right3),
        (slice_right0, slice_right2, slice_right3),
        (slice_right0, slice_right3),
    ]

    for ls in left_slices:
        for rs in right_slices:
            sketched_tt_block = blocked_stream_sketch(
                X_tensor, left_drm, right_drm, ls, rs
            )
            for Y1, Y2 in zip(
                sketched_tt_block.Psi_cores, sketched_tt.Psi_cores
            ):
                assert np.allclose(Y1, Y2)
            for Z1, Z2 in zip(
                sketched_tt_block.Omega_mats, sketched_tt.Omega_mats
            ):
                assert np.allclose(Z1, Z2)


def general_to_hmt(
    tensor,
    left_rank,
    right_rank,
    seed=None,
    left_drm_type=None,
    right_drm_type=None,
):
    return hmt_sketch(tensor, right_rank, seed, right_drm_type)


SKETCH_FUNC_DISPATCHER = {
    SketchMethod.hmt: general_to_hmt,
    SketchMethod.streaming: stream_sketch,
    SketchMethod.orthogonal: orthogonal_sketch,
}


def general_exact_recovery(
    X_dense,
    X_tensor,
    left_rank,
    right_rank,
    seed,
    left_drm_type,
    right_drm_type,
    sketch_method: SketchMethod,
):
    sketch_func = SKETCH_FUNC_DISPATCHER[sketch_method]
    tt_sketched = sketch_func(
        X_tensor,
        left_rank,
        right_rank,
        seed=seed,
        left_drm_type=left_drm_type,
        right_drm_type=right_drm_type,
    )
    error = tt_sketched.error(X_dense)
    tt_sketched_dense = tt_sketched.to_numpy()
    assert error < 1e-9

    tt_sketched2 = sketch_func(
        X_tensor,
        left_rank,
        right_rank,
        seed=seed,
        left_drm_type=left_drm_type,
        right_drm_type=right_drm_type,
    )
    # same seed should give exact same result. But in the case of parallel
    # execution, we can sum in different order giving same result only up to
    # machine epsilon
    assert tt_sketched2.error(tt_sketched) < 1e-9

    tt_sketched3 = sketch_func(
        X_tensor,
        left_rank,
        right_rank,
        seed=seed + 1,
        left_drm_type=left_drm_type,
        right_drm_type=right_drm_type,
    )
    tt_sketched_dense3 = tt_sketched3.to_numpy()
    # different seed should give different result
    assert not np.all(tt_sketched_dense == tt_sketched_dense3)


sparse_drm_list = [
    drm_type.__name__
    for drm_type in ALL_DRM
    if issubclass(drm_type, CansketchSparse)
]
sparse_drm_pairs = [
    "|".join(s) for s in itertools.permutations(sparse_drm_list, 2)
]
sparse_drm_pairs.extend(["|".join([s, s]) for s in sparse_drm_list])
assert len(sparse_drm_pairs) > 0


@pytest.mark.parametrize("drm_types", sparse_drm_pairs)
@pytest.mark.parametrize("n_dims", [2, 3])
@pytest.mark.parametrize("rank", [2, 5])
@pytest.mark.parametrize("sketch_method", AVAILABLE_SKETCHING_METHODS)
def test_exact_recovery_sparse(n_dims, rank, drm_types, sketch_method):
    seed = 180
    X_shape = tuple(range(9, 9 + n_dims))
    X_tt = TensorTrain.random(X_shape, rank)
    X = X_tt.to_numpy()

    left_rank = tuple(range(rank, rank + n_dims - 1))
    right_rank = tuple(range(rank + 1, rank + n_dims))

    X_sparse = X_tt.dense().to_sparse()

    drm_types = drm_types.split("|")
    left_drm = DRM_DICT[drm_types[0]]
    right_drm = DRM_DICT[drm_types[1]]
    general_exact_recovery(
        X,
        X_sparse,
        left_rank,
        right_rank,
        seed,
        left_drm,
        right_drm,
        sketch_method,
    )

    if issubclass(left_drm, CanSlice) and issubclass(right_drm, CanSlice):
        general_blocked_sketch(X_sparse, seed, left_drm, right_drm)

    if issubclass(left_drm, CanIncreaseRank) and issubclass(
        right_drm, CanIncreaseRank
    ):
        general_rank_increase(
            X, X_sparse, left_rank, right_rank, seed, left_drm, right_drm
        )


@pytest.mark.parametrize("direction", ["left", "right"])
def test_massive_oversample(direction):
    n_dims = 4
    rank = 5
    seed = 180
    X_shape = tuple(range(5, 5 + n_dims))
    X_tt = TensorTrain.random(X_shape, rank)
    X = X_tt.to_numpy()
    if direction == "left":
        left_rank = 100
        right_rank = 90
    if direction == "right":
        left_rank = 90
        right_rank = 100
    stt = stream_sketch(
        X_tt,
        left_rank,
        right_rank,
        seed=seed,
    )

    left_tt = TensorTrain(stt.C_cores(direction="left"))
    right_tt = TensorTrain(stt.C_cores(direction="right"))

    assert np.allclose(left_tt.to_numpy(), right_tt.to_numpy())
    assert left_tt.rank == stt.right_rank
    assert right_tt.rank == stt.left_rank


@pytest.mark.parametrize("n_dims", [2, 3, 4])
def test_sketch_dense(n_dims):
    seed = 179
    shape = (10,) * n_dims
    left_rank = 100
    right_rank = 200
    tensor_data = np.random.normal(size=shape)
    tensor = DenseTensor(tensor_data)
    stt = stream_sketch(
        tensor,
        left_rank,
        right_rank,
        seed=seed,
        left_drm_type=DenseGaussianDRM,
        right_drm_type=DenseGaussianDRM,
    )
    assert np.linalg.norm(stt.to_numpy() - tensor.to_numpy()) < 1e-9
    stt = orthogonal_sketch(
        tensor,
        left_rank,
        right_rank,
        seed=seed,
        left_drm_type=DenseGaussianDRM,
        right_drm_type=DenseGaussianDRM,
    )
    assert np.linalg.norm(stt.to_numpy() - tensor.to_numpy()) < 1e-9


@pytest.mark.parametrize("sketch_method", AVAILABLE_SKETCHING_METHODS)
def test_tensor_sum(sketch_method):
    seed = 179
    rank = 4
    n_dims = 4

    X_shape = tuple(range(7, 7 + n_dims))

    X1_tt = TensorTrain.random(X_shape, rank)
    X1 = X1_tt.to_numpy()

    left_rank = tuple(range(rank, rank + n_dims - 1))
    right_rank = tuple(range(rank + 1, rank + n_dims))

    X_sparse = X1_tt.dense().to_sparse()
    X_sparse_sum16 = X_sparse.split(16)
    X_sparse_sum2 = X_sparse.split(2)
    assert isinstance(X_sparse_sum2, TensorSum)
    assert isinstance(X_sparse_sum16, TensorSum)

    left_drm_type = TensorTrainDRM
    right_drm_type = SparseGaussianDRM

    # Recovery should work for TensorSum
    general_exact_recovery(
        X1,
        X_sparse_sum16,
        left_rank,
        right_rank,
        seed,
        left_drm_type,
        right_drm_type,
        sketch_method,
    )
    general_exact_recovery(
        X1,
        X_sparse_sum2,
        left_rank,
        right_rank,
        seed,
        left_drm_type,
        right_drm_type,
        sketch_method,
    )
    # Sketching is linear, so doing it as (parallel) sum should give same result
    stt1 = stream_sketch(X_sparse_sum16, left_rank, right_rank, seed)
    stt3 = stream_sketch(X_sparse_sum2, left_rank, right_rank, seed)
    stt2 = stream_sketch(X_sparse, left_rank, right_rank, seed)
    for Y1, Y2, Y3 in zip(stt1.Psi_cores, stt2.Psi_cores, stt3.Psi_cores):
        assert np.allclose(Y1, Y2)
        assert np.allclose(Y1, Y3)
    for Y1, Y2, Y3 in zip(stt1.Omega_mats, stt2.Omega_mats, stt3.Omega_mats):
        assert np.allclose(Y1, Y2)
        assert np.allclose(Y1, Y3)

    X2_tt = TensorTrain.random(X_shape, rank)
    X2 = X1 + X2_tt.to_numpy()

    X1_plus_X2 = X_sparse + X2_tt
    left_drm_type = TensorTrainDRM
    right_drm_type = TensorTrainDRM
    left_rank_bigger = tuple(r + 10 for r in left_rank)
    right_rank_bigger = tuple(r + 10 for r in right_rank)
    stt4 = stream_sketch(
        X1_plus_X2,
        left_rank_bigger,
        right_rank_bigger,
        seed,
        left_drm_type=left_drm_type,
        right_drm_type=right_drm_type,
    )
    assert stt4.to_tt().error(X2) < 1e-8

    # test the user-friendlty way of updating sketch
    stt5 = stream_sketch(
        X_sparse,
        left_rank_bigger,
        right_rank_bigger,
        seed,
        left_drm_type=left_drm_type,
        right_drm_type=right_drm_type,
    )
    stt6 = stt5 + X2_tt
    assert stt6.error(X2) < 1e-8


tt_drm_list = [
    drm_type.__name__
    for drm_type in ALL_DRM
    if issubclass(drm_type, CansketchTT)
]
tt_drm_pairs = ["|".join(s) for s in itertools.permutations(tt_drm_list, 2)]
tt_drm_pairs.extend(["|".join([s, s]) for s in tt_drm_list])
assert len(tt_drm_pairs) > 0


@pytest.mark.parametrize("drm_types", tt_drm_pairs)
@pytest.mark.parametrize("n_dims", [2, 3])
@pytest.mark.parametrize("rank", [2, 3])
@pytest.mark.parametrize("sketch_method", AVAILABLE_SKETCHING_METHODS)
def test_exact_recovery_tt(n_dims, rank, drm_types, sketch_method):
    seed = 180
    X_shape = tuple(range(10, 10 + n_dims))
    X_tt = TensorTrain.random(X_shape, rank, seed=seed)
    X = X_tt.to_numpy()

    left_rank = tuple(range(rank, rank + n_dims - 1))
    right_rank = tuple(range(rank + 1, rank + n_dims))

    drm_types = drm_types.split("|")
    left_drm_type = DRM_DICT[drm_types[0]]
    right_drm_type = DRM_DICT[drm_types[1]]
    general_exact_recovery(
        X,
        X_tt,
        left_rank,
        right_rank,
        seed,
        left_drm_type=left_drm_type,
        right_drm_type=right_drm_type,
        sketch_method=sketch_method,
    )

    if issubclass(left_drm_type, CanSlice) and issubclass(
        right_drm_type, CanSlice
    ):
        general_blocked_sketch(
            X_tt,
            seed,
            left_drm_type=left_drm_type,
            right_drm_type=right_drm_type,
        )

    if issubclass(left_drm_type, CanIncreaseRank) and issubclass(
        right_drm_type, CanIncreaseRank
    ):
        general_rank_increase(
            X,
            X_tt,
            left_rank,
            right_rank,
            seed,
            left_drm_type=left_drm_type,
            right_drm_type=right_drm_type,
        )


cp_drm_list = [
    drm_type.__name__
    for drm_type in ALL_DRM
    if issubclass(drm_type, CansketchCP)
]
cp_drm_pairs = ["|".join(s) for s in itertools.permutations(cp_drm_list, 2)]
cp_drm_pairs.extend(["|".join([s, s]) for s in cp_drm_list])
assert len(cp_drm_pairs) > 0


@pytest.mark.parametrize("n_dims", [2, 3])
@pytest.mark.parametrize("rank", [2, 3])
@pytest.mark.parametrize("sketch_method", AVAILABLE_SKETCHING_METHODS)
def test_exact_recovery_tucker(n_dims, rank, sketch_method):
    seed = 180
    X_shape = tuple(range(10, 10 + n_dims))
    X_tucker = TuckerTensor.random(X_shape, rank, seed=seed)
    X = X_tucker.to_numpy()

    left_rank = tuple(range(rank, rank + n_dims - 1))
    right_rank = tuple(range(rank + 1, rank + n_dims))

    general_exact_recovery(
        X,
        X_tucker,
        left_rank,
        right_rank,
        seed,
        left_drm_type=TensorTrainDRM,
        right_drm_type=TensorTrainDRM,
        sketch_method=sketch_method,
    )


@pytest.mark.parametrize("drm_types", cp_drm_pairs)
@pytest.mark.parametrize("n_dims", [2, 3])
@pytest.mark.parametrize("rank", [2, 3])
@pytest.mark.parametrize("sketch_method", AVAILABLE_SKETCHING_METHODS)
def test_exact_recovery_cp(n_dims, rank, drm_types, sketch_method):
    seed = 180
    X_shape = tuple(range(10, 10 + n_dims))
    X_cp = CPTensor.random(X_shape, rank, seed=seed)
    X = X_cp.to_numpy()

    left_rank = tuple(range(rank, rank + n_dims - 1))
    right_rank = tuple(range(rank + 1, rank + n_dims))

    drm_types = drm_types.split("|")
    left_drm_type = DRM_DICT[drm_types[0]]
    right_drm_type = DRM_DICT[drm_types[1]]
    general_exact_recovery(
        X,
        X_cp,
        left_rank,
        right_rank,
        seed,
        left_drm_type=left_drm_type,
        right_drm_type=right_drm_type,
        sketch_method=sketch_method,
    )

    if issubclass(left_drm_type, CanSlice) and issubclass(
        right_drm_type, CanSlice
    ):
        general_blocked_sketch(
            X_cp,
            seed,
            left_drm_type=left_drm_type,
            right_drm_type=right_drm_type,
        )

    if issubclass(left_drm_type, CanIncreaseRank) and issubclass(
        right_drm_type, CanIncreaseRank
    ):
        general_rank_increase(
            X,
            X_cp,
            left_rank,
            right_rank,
            seed,
            left_drm_type=left_drm_type,
            right_drm_type=right_drm_type,
        )


# TODO: do this test for all DRM's
@pytest.mark.parametrize("n_dims", [2, 4])
@pytest.mark.parametrize("rank", [2, 4])
@pytest.mark.parametrize("right_rank_oversample", [1, 2])
def test_sketch_linearity(n_dims, rank, right_rank_oversample):
    N = 1000
    shape = (10,) * n_dims
    left_rank = (rank,) * (n_dims - 1)
    right_rank = (rank + right_rank_oversample,) * (n_dims - 1)
    tot_dim = np.prod(shape)
    N = min(N, tot_dim)

    inds_flat = np.random.choice(tot_dim, size=N)
    inds = np.stack(np.unravel_index(inds_flat, shape))

    entries = np.random.normal(size=N)

    X_sparse = SparseTensor(shape, inds, entries)
    X_sparse1 = SparseTensor(shape, inds[:, : N // 2], entries[: N // 2])
    X_sparse2 = SparseTensor(shape, inds[:, N // 2 :], entries[N // 2 :])
    seed = 179
    tt_sketched = stream_sketch(X_sparse, left_rank, right_rank, seed=seed)
    tt_sketched1 = stream_sketch(X_sparse1, left_rank, right_rank, seed=seed)
    tt_sketched2 = stream_sketch(X_sparse2, left_rank, right_rank, seed=seed)
    for mu in range(n_dims):
        assert np.allclose(
            tt_sketched1.Psi_cores[mu] + tt_sketched2.Psi_cores[mu],
            tt_sketched.Psi_cores[mu],
        )
    for mu in range(n_dims - 1):
        assert np.allclose(
            tt_sketched1.Omega_mats[mu] + tt_sketched2.Omega_mats[mu],
            tt_sketched.Omega_mats[mu],
        )


@pytest.mark.parametrize("n_dims", [3, 4])
@pytest.mark.parametrize("rank", [2, 4])
@pytest.mark.parametrize("bigger_side", ["left", "right"])
def test_tt_cores_contraction(n_dims, rank, bigger_side):
    seed = 180
    X_shape = tuple(range(5, 5 + n_dims))
    rank = tuple(range(rank, rank + n_dims - 1)[::-1])
    tt = TensorTrain.random(X_shape, rank)

    if bigger_side == "left":
        left_rank = [r + 2 for r in tt.rank]
        right_rank = [r + 1 for r in tt.rank]
    if bigger_side == "right":
        left_rank = [r + 1 for r in tt.rank]
        right_rank = [r + 2 for r in tt.rank]

    stt = stream_sketch(tt, left_rank, right_rank, seed=seed)

    assert (np.linalg.norm(tt.to_numpy() - stt.to_tt().to_numpy())) < 1e-8
    left_tt = TensorTrain(stt.C_cores(direction="left"))
    right_tt = TensorTrain(stt.C_cores(direction="right"))

    assert np.allclose(left_tt.to_numpy(), right_tt.to_numpy())
    assert left_tt.rank == stt.right_rank
    assert right_tt.rank == stt.left_rank


def test_defaults_sketch():
    """Test that the default settings are correct."""
    shape = (8, 9, 10)
    r = 10
    test_tensors = []

    test_tensor_tt = TensorTrain.random(shape, r)
    test_tensors.append(test_tensor_tt)

    test_tensor_sparse = SparseTensor.random(shape, nnz=10)
    test_tensors.append(test_tensor_sparse)

    test_tensor_tucker = TuckerTensor.random(shape, 2)
    test_tensors.append(test_tensor_tucker)

    test_tensor_dense = test_tensor_sparse.dense()
    test_tensors.append(test_tensor_dense)

    test_tensor_sum = test_tensor_tt + test_tensor_sparse
    test_tensors.append(test_tensor_sum)

    test_tensor_cp = CPTensor.random(shape, 5)
    test_tensors.append(test_tensor_cp)

    for test_tensor in test_tensors:
        sketched = hmt_sketch(test_tensor, r + 1)
        assert sketched.error(test_tensor) < 1e-8
        sketched = orthogonal_sketch(
            test_tensor, left_rank=r + 1, right_rank=r + 2
        )
        assert sketched.error(test_tensor) < 1e-8
        sketched = stream_sketch(test_tensor, left_rank=r + 1, right_rank=r + 2)
        assert sketched.error(test_tensor) < 1e-8


def test_sketched_tt_arithmetic():
    n_dims = 3
    rank = tuple(range(3, 3 + n_dims - 1))[::-1]
    right_rank = tuple(r + 1 for r in rank)
    X_shape = tuple(range(9, 9 + n_dims))
    X_tt = TensorTrain.random(X_shape, rank)
    tts1 = stream_sketch(X_tt, left_rank=rank, right_rank=right_rank)
    tt1 = tts1.to_tt()

    tts2 = tts1 * 2
    tt2 = tts2.to_tt()

    assert tt2.error(tt1 * 2) < 1e-10

    tts3 = tts1.T
    tt3 = tts3.to_tt()
    assert tt3.error(tt1.T) < 1e-10


"""Implements various types of tensors"""
from __future__ import annotations

from abc import ABC, abstractmethod, abstractproperty
from copy import deepcopy
from dataclasses import dataclass
from functools import cached_property
from typing import (
    Iterable,
    List,
    Optional,
    Sequence,
    Dict,
    Tuple,
    Union,
    TypeVar,
    Generic,
)
import warnings

import numpy as np
from numpy.random import SeedSequence
import numpy.typing as npt

from tt_sketch.utils import ArrayList, TTRank, process_tt_rank, random_normal

TType = TypeVar("TType", bound="Tensor")


class Tensor(ABC):
    """Abstract base class for tensors."""

    #: The shape of the tensor
    shape: Tuple[int, ...]

    @abstractproperty
    def T(self: TType) -> TType:
        """Transpose of the tensor.

        If a tensor has shape (n1,n2,...,nd), then transpose of the tensor
        has shape (nd,...,n2,n1).
        """

    @abstractproperty
    def size(self) -> int:
        """Number of floating point elements used to store the tensor."""

    @abstractmethod
    def to_numpy(self) -> npt.NDArray[np.float64]:
        """Converts the tensor to a (dense) numpy array of same shape."""

    def error(
        self: TType,
        other: TType,
        relative: bool = False,
        rmse: bool = False,
        fast: bool = False,
    ) -> float:
        """L2 error of the tensor.

        If ``fast=True``, then error is computed using inner product formula.
        This is not numerically stable, and gives inaccurate results below
        relative errors of around 1e-8.
        """
        if isinstance(other, np.ndarray):
            other = DenseTensor(other)
        other_norm = other.norm()
        if fast:
            self_norm = self.norm()
            dot = self.dot(other)
            norm_sum = self_norm**2 + other_norm**2
            error = np.sqrt(norm_sum) * np.sqrt(np.abs(1 - 2 * dot / norm_sum))
        else:
            error = np.linalg.norm(self.to_numpy() - other.to_numpy())
        if relative:
            if other_norm == 0:
                return np.inf
            error /= other_norm
        if rmse:
            error /= np.sqrt(np.prod(self.shape))
        return error

    @property
    def ndim(self) -> int:
        """Number of modes of the tensor."""
        return len(self.shape)

    def dense(self) -> DenseTensor:
        """Converts to ``DenseTensor`` object"""
        dense_tensor = self.to_numpy()
        return DenseTensor(dense_tensor)

    def __add__(self, other) -> TensorSum:
        """Addition of two tensors produces ``TensorSum`` object"""
        if isinstance(other, TensorSum):
            if not isinstance(self, TensorSum):
                return TensorSum([self] + other.tensors)
            else:
                return TensorSum(self.tensors + other.tensors)
        elif isinstance(self, TensorSum):
            return TensorSum(self.tensors + [other])
        else:
            return TensorSum([self, other])

    @abstractmethod
    def __mul__(self: TType, other: float) -> TType:
        """Multiplication by a scalar"""

    def __rmul__(self: TType, other: float) -> TType:
        return self.__mul__(other)

    def __truediv__(self, other: float):
        return self.__mul__(1 / other)

    def __sub__(self, other: Tensor) -> Tensor:
        return self + (-other)

    def __neg__(self) -> Tensor:
        return self * -1

    def dot(self: TType, other: TType, reverse=False) -> float:
        """Dot product of two tensors"""
        if isinstance(other, TensorSum):
            return other.dot(self)
        if not reverse:  # try first to see if other can dot self
            return other.dot(self, reverse=True)
        self_np = self.to_numpy().reshape(-1)
        other_np = other.to_numpy().reshape(-1)
        return np.dot(self_np, other_np)

    def norm(self) -> float:
        """L2 norm of the tensor"""
        # np.abs because dot can be negative due to numerical errors
        return np.sqrt(np.abs(self.dot(self)))

    def __matmul__(self: TType, other: TType) -> float:
        return self.dot(other)


class DenseTensor(Tensor):
    shape: Tuple[int, ...]
    data: npt.NDArray[np.float64]

    def __init__(self, data: npt.NDArray) -> None:
        self.shape = data.shape
        self.data = data

    def to_sparse(self) -> SparseTensor:
        """
        Converts to a sparse tensor.

        This is mainly used for testing sketching algorithms, as there is
        otherwise no reason to convert a dense tensor to a sparse tensor.
        """
        X = self.data
        n_dims = len(X.shape)
        inds = np.indices(X.shape).reshape(n_dims, -1)
        entries = X.reshape(-1)
        return SparseTensor(X.shape, inds, entries)

    @property
    def T(self) -> DenseTensor:
        permutation = tuple(range(len(self.shape))[::-1])
        new_data = np.transpose(self.data, permutation)
        return self.__class__(new_data)

    @property
    def size(self) -> int:
        return np.prod(self.shape)

    def to_numpy(self) -> npt.NDArray[np.float64]:
        return self.data

    def __repr__(self) -> str:
        return f"<Dense tensor of shape {self.shape} at {hex(id(self))}>"

    def __mul__(self, other: float) -> DenseTensor:
        return self.__class__(self.data * other)

    @classmethod
    def random(cls, shape: Tuple[int, ...]) -> DenseTensor:
        return cls(random_normal(shape))


@dataclass
class SparseTensor(Tensor):
    #: The shape of the tensor
    shape: Tuple[int, ...]

    #: The indices of the non-zero entries
    indices: Union[npt.NDArray, Tuple[npt.NDArray, ...]]

    #: The entries of the non-zero entries
    entries: npt.NDArray[np.float64]

    def __post_init__(self):
        if isinstance(self.indices, tuple):
            self.indices = np.stack(self.indices)

    @property
    def T(self) -> SparseTensor:
        new_indices = self.indices[::-1]
        new_shape = self.shape[::-1]
        return self.__class__(new_shape, new_indices, self.entries)

    @property
    def size(self) -> int:
        return self.nnz * (self.ndim + 1)

    @property
    def nnz(self) -> int:
        """Number of non-zero entries."""
        return len(self.entries)

    def split(self, n_summands: int) -> TensorSum:
        """
        Splits the tensor a ``TensorSum`` containing ``n_summands`` tensors.
        """
        block_size = self.nnz // n_summands
        summands: List[Tensor] = []
        for i in range(n_summands):
            if i < n_summands - 1:
                summand_slice = slice(i * block_size, (i + 1) * block_size)
            else:
                summand_slice = slice(i * block_size, self.nnz)
            summand_indices = tuple(ind[summand_slice] for ind in self.indices)
            summands.append(
                SparseTensor(
                    self.shape,
                    summand_indices,
                    self.entries[summand_slice],
                )
            )
        return TensorSum(summands)

    def to_numpy(self) -> npt.NDArray[np.float64]:
        X = np.zeros(self.shape)
        X[tuple(self.indices)] = self.entries
        return X

    def norm(self) -> float:
        return float(np.linalg.norm(self.entries))

    def __repr__(self) -> str:
        return (
            f"<Sparse tensor of shape {self.shape} with {self.nnz} non-zero"
            f" entries at {hex(id(self))}>"
        )

    def dot(self, other: Tensor, reverse=False) -> float:
        try:
            other_entries = other.gather(self.indices)
            return np.dot(other_entries, self.entries)
        except AttributeError:
            return super().dot(other, reverse=reverse)

    @classmethod
    def random(
        cls, shape: Tuple[int, ...], nnz: int, seed: Optional[int] = None
    ) -> SparseTensor:
        """
        Generates a random sparse tensor with `nnz` non-zero gaussian entries
        """
        if seed is not None:
            np.random.seed(seed)
        total_size = np.prod(shape)
        indices_flat = np.random.choice(total_size, size=nnz, replace=False)
        indices = np.unravel_index(indices_flat, shape)
        entries = random_normal(shape=(nnz,), seed=seed)
        return cls(shape, indices, entries)

    def __mul__(self, other: float) -> SparseTensor:
        return self.__class__(self.shape, self.indices, self.entries * other)

    def gather(
        self, indices: Tuple[npt.NDArray, ...]
    ) -> npt.NDArray[np.float64]:
        """
        Gathers the entries corresponding to the given indices.
        """
        dense_indices = np.ravel_multi_index(indices, self.shape)
        out = np.zeros(len(dense_indices))
        dic = self.dict
        for i, ind in enumerate(dense_indices):
            out[i] = dic.get(ind, 0.0)
        return out

    @cached_property
    def dict(self) -> Dict[int, float]:
        dense_indices = np.ravel_multi_index(self.indices, self.shape)
        return {i: v for i, v in zip(dense_indices, self.entries)}


class TensorTrain(Tensor):

    #: The shape of the tensor
    shape: Tuple[int, ...]

    #: Tuple encoding the tensor train rank
    rank: Tuple[int, ...]

    #: A list containing the cores of the tensor train
    cores: ArrayList

    def __init__(self, cores: ArrayList) -> None:
        self.cores = cores
        self.shape = tuple(C.shape[1] for C in cores)
        self.rank = tuple(C.shape[0] for C in cores[1:])

    @property
    def T(self) -> TensorTrain:
        new_cores = [np.transpose(C, (2, 1, 0)) for C in self.cores[::-1]]
        return self.__class__(new_cores)

    def to_numpy(self) -> npt.NDArray:
        dense_tensor = self.cores[0]
        dense_tensor = dense_tensor.reshape(dense_tensor.shape[1:])
        for C in self.cores[1:]:
            dense_tensor = np.einsum("...j,jkl->...kl", dense_tensor, C)
        dense_tensor = dense_tensor.reshape(dense_tensor.shape[:-1])
        return dense_tensor

    @classmethod
    def random(
        cls,
        shape: Tuple[int, ...],
        rank: TTRank,
        seed: Optional[int] = None,
        orthog: bool = False,
        trim: Optional[bool] = None,
        norm_goal: str = "norm-1",
    ) -> TensorTrain:
        """
        Generate random orthogonal tensor train cores

        By default, a core of ``(r1, n, r2)`` has Gaussian entries with zero
        mean and variance ``1 / r1 * n * r2``, so that expected Frobenius norm
        is 1.

        If ``trim=True``, the ranks are trimmed; i.e. we enforce that r1*n>=r2
        and r2*n>=r1a.

        If ``orthog`` is set to ``True``, all cores except the last are
        left-orthogonalized. Trim must be enabled in this case.
        """

        d = len(shape)
        if trim is None:
            trim = True if orthog else False
        if orthog and not trim:
            raise ValueError(
                "Trimming must be enabled if orthogonalization is enabled."
            )
        rank = process_tt_rank(rank, shape, trim=trim)
        rank_augmented = (1,) + tuple(rank) + (1,)

        cores = []
        seq = SeedSequence(seed)
        seeds = seq.generate_state(d)
        for i in range(d):
            r1 = rank_augmented[i]
            r2 = rank_augmented[i + 1]
            n = shape[i]
            core = random_normal(shape=(r1 * n, r2), seed=seeds[i])

            if orthog and i < d - 1:
                core, _ = np.linalg.qr(core, mode="reduced")
            elif norm_goal == "norm-1":
                core /= np.sqrt(r1 * n)
            elif norm_goal == "norm-preserve":
                core /= np.sqrt(r1)
            else:
                raise ValueError(f"Unknown norm goal: {norm_goal}")

            core = core.reshape(r1, n, r2)
            cores.append(core)

        return cls(cores)

    @classmethod
    def zero(cls, shape: Tuple[int, ...], rank: TTRank) -> TensorTrain:
        d = len(shape)

        rank = process_tt_rank(rank, shape, trim=False)
        cores = []
        for (r1, d, r2) in zip((1,) + rank, shape, rank + (1,)):
            cores.append(np.zeros((r1, d, r2)))
        return cls(cores)

    def partial_dense(self, dir: str = "lr") -> ArrayList:
        """Do partial contractions to dense tensor; ``X[0].X[1]...X[mu]``"""
        if dir == "lr":
            partial_cores = [self.cores[0].reshape(-1, self.cores[0].shape[-1])]
            for core in self.cores[1:-1]:
                new_core = np.einsum("ij,jkl->ikl", partial_cores[-1], core)
                new_core = new_core.reshape(-1, new_core.shape[-1])
                partial_cores.append(new_core)
        elif dir == "rl":
            partial_cores = [
                self.cores[-1].reshape(self.cores[-1].shape[0], -1)
            ]
            for core in self.cores[-2:0:-1]:
                new_core = np.einsum("ijk,kl->ijl", core, partial_cores[-1])
                new_core = new_core.reshape(new_core.shape[0], -1)
                partial_cores.append(new_core)
        return partial_cores

    def __getitem__(self, index: int) -> npt.NDArray:
        return self.cores[index]

    def __setitem__(self, index: int, data: npt.NDArray) -> None:
        self.cores[index] = data

    def gather(
        self,
        idx: Union[npt.NDArray[np.int64], Tuple[npt.NDArray[np.int64], ...]],
    ) -> npt.NDArray:
        """Gather entries of dense tensor according to indices.

        For each row of `idx` this returns one number. This number is obtained
        by multiplying the slices of each core corresponding to each index (in
        a left-to-right fashion).
        """
        if not isinstance(idx, np.ndarray):
            idx_array = np.stack(idx)
        else:
            idx_array = np.array(idx)
        N = idx_array.shape[1]
        result = np.take(
            self[0].reshape(self[0].shape[1:]), idx_array[0], axis=0
        )
        for i in range(1, self.ndim):
            r = self[i].shape[2]
            next_step = np.zeros((N, r))
            for j in range(self.shape[i]):
                idx_mask = np.where(idx_array[i] == j)
                mat = self[i][:, j, :]
                next_step[idx_mask] = result[idx_mask] @ mat
            result = next_step
        return result.reshape(-1)

    def norm(self) -> float:
        self_orth = self.orthogonalize()
        return np.linalg.norm(self_orth.cores[-1])

    def round(
        self,
        eps: Optional[float] = None,
        max_rank: Optional[TTRank] = None,
        orthogonalized: bool = False,
    ) -> TensorTrain:
        """Standard TT-SVD rounding scheme.

        First left orthogonalize in LR sweep, then apply SVD-based rounding in a
        RL sweep. Leaves the tensor right-orthogonalized.

        If the tensor is already orthogonalized, then pass
        ``orthogonalized=True`` to avoid unnecessary re-orthogonalization."""
        if not orthogonalized:
            tt = self.orthogonalize()  # left-orthogonalize
        else:
            tt = self
        if eps is None:
            eps = 0
        if max_rank is None:
            max_rank = tt.rank
        max_rank = process_tt_rank(max_rank, tt.shape, trim=True)

        new_cores = []
        US_trunc: npt.NDArray
        for mu, C in list(enumerate(tt.cores))[::-1]:
            if mu < tt.ndim - 1:
                C = np.einsum("ijk,kl->ijl", C, US_trunc)
            if mu > 0:
                C_mat = C.reshape(C.shape[0], C.shape[1] * C.shape[2])
                U, S, Vt = np.linalg.svd(C_mat)
                r = max(1, min(np.sum(S > S[0] * eps), max_rank[mu - 1]))
                US_trunc = U[:, :r] @ np.diag(S[:r])
                Vt_trunc = Vt[:r, :]
                new_cores.append(Vt_trunc.reshape(r, C.shape[1], C.shape[2]))
            else:
                new_cores.append(C)

        return self.__class__(new_cores[::-1])

    def svdvals(self) -> List[npt.NDArray]:
        """Return singular value of each mode"""
        tt = self.orthogonalize()
        svdvals = []
        U: npt.NDArray
        S: npt.NDArray
        for mu, C in list(enumerate(tt.cores))[::-1]:
            if mu < tt.ndim - 1:
                US = U @ np.diag(S)
                C = np.einsum("ijk,kl->ijl", C, US)

            if mu > 0:
                C_mat = C.reshape(C.shape[0], C.shape[1] * C.shape[2])
            else:
                C_mat = C.reshape(C.shape[0] * C.shape[1], C.shape[2])

            U, S, _ = np.linalg.svd(C_mat)
            svdvals.append(S)
        svdvals = svdvals[::-1]
        return svdvals

    def __mul__(self, other: float) -> TensorTrain:
        new_cores = deepcopy(self.cores)
        # Absorb number into last core, since if we do orthogonalize, we left
        # orthogonalize.
        new_cores[-1] = new_cores[-1] * other
        return self.__class__(new_cores)

    __rmul__ = __mul__

    @property
    def size(self) -> int:
        return sum(core.size for core in self.cores)

    def add(self, other: TensorTrain) -> TensorTrain:
        """
        Add two ``TensorTrain`` by taking direct sums of cores.

        Note, this does not overload the addition operator ``+``; the addition
        operator instead returns a lazy ``TensorSum`` object.
        """
        new_cores = [np.concatenate((self.cores[0], other.cores[0]), axis=2)]
        for C1, C2 in zip(self.cores[1:-1], other.cores[1:-1]):
            r1, d, r2 = C1.shape
            r3, _, r4 = C2.shape
            zeros1 = np.zeros((r1, d, r4))
            zeros2 = np.zeros((r3, d, r2))
            row1 = np.concatenate((C1, zeros1), axis=2)
            row2 = np.concatenate((zeros2, C2), axis=2)
            new_cores.append(np.concatenate((row1, row2), axis=0))
        new_cores.append(
            np.concatenate((self.cores[-1], other.cores[-1]), axis=0)
        )
        new_tt = self.__class__(new_cores)
        return new_tt

    def dot(self, other: Tensor, reverse=False) -> float:
        """
        Compute the dot product of two tensor trains with the same shape.

        Result is computed in a left-to-right sweep.
        """
        if isinstance(other, TensorTrain):
            result = np.einsum("ijk,ljm->km", self.cores[0], other.cores[0])
            for core1, core2 in zip(self.cores[1:], other.cores[1:]):
                # optimize reduces complexity from r^4*n to r^3*n
                result = np.einsum(
                    "ij,ika,jkb->ab", result, core1, core2, optimize="optimal"
                )
            return np.sum(result)
        else:
            return super().dot(other, reverse=reverse)

    def orthogonalize(self) -> TensorTrain:
        """Do QR sweep to left-orthogonalize"""
        new_cores = []
        R: npt.NDArray
        for mu, C in enumerate(self.cores):
            if mu > 0:
                C = np.einsum("ij,jkl->ikl", R, C)
            if mu < self.ndim - 1:
                C_mat = C.reshape(C.shape[0] * C.shape[1], C.shape[2])
                Q, R = np.linalg.qr(C_mat)
                new_cores.append(Q.reshape(C.shape[0], C.shape[1], -1))
            else:
                new_cores.append(C)
        return self.__class__(new_cores)

    def __repr__(self) -> str:
        return (
            f"<Tensor train of shape {self.shape} with rank {self.rank}"
            f" at {hex(id(self))}>"
        )

    def error(
        self: TType,
        other: TType,
        relative: bool = False,
        rmse: bool = False,
        fast: bool = False,
    ) -> float:
        """
        L2 error of tensor, see docs for ``Tensor::error``.

        Overloads only the error between two Tensor Trains
        using much faster accurate method.
        """
        try:  # Try coercing to TensorTrain
            other = other.to_tt()
        except AttributeError:
            pass

        if isinstance(other, TensorTrain):
            error = self.add(-other).norm()
            if relative:
                other_norm = other.norm()
                if other_norm == 0:
                    return np.inf
                error /= other_norm
            if rmse:
                error /= np.sqrt(np.prod(self.shape))
            return error
        else:
            return super().error(other, relative=relative, rmse=rmse, fast=fast)


class TensorSum(Generic[TType], Tensor):
    """Container for sums of tensors"""

    shape: Tuple[int, ...]
    tensors: List[TType]

    def __init__(self, tensors: List[TType], shape=None) -> None:
        if shape is None:
            shape = tensors[0].shape
        self.shape = shape
        self.tensors = tensors

    @property
    def size(self) -> int:
        return sum(tensor.size for tensor in self.tensors)

    @property
    def T(self) -> TensorSum:
        new_tensors: List[TType] = [X.T for X in self.tensors]
        return self.__class__(new_tensors)

    def to_numpy(self) -> npt.NDArray[np.float64]:
        s = np.zeros(self.shape)
        for X in self.tensors:
            s += X.to_numpy()
        return s

    def __add__(self, other: TType) -> TensorSum:
        if isinstance(other, TensorSum):
            return self.__class__(self.tensors + other.tensors)
        else:
            return self.__class__(self.tensors + [other])

    def __iadd__(self, other: TType) -> TensorSum:
        if isinstance(other, TensorSum):
            self.tensors.extend(other.tensors)
        else:
            self.tensors.append(other)
        return self

    def __repr__(self) -> str:
        return (
            f"<Sum of {self.num_summands} tensors of shape {self.shape}"
            f" at {hex(id(self))}>"
        )

    @property
    def num_summands(self) -> int:
        return len(self.tensors)

    def __mul__(self, other: Union[float, Iterable[float]]) -> TensorSum:
        try:
            return self.__class__(
                [X * c for X, c in zip(self.tensors, other, strict=True)]
            )
        except TypeError:
            return self.__class__([X * other for X in self.tensors])

    def dot(self, other: Tensor, reverse=False) -> float:
        return sum(X.dot(other, reverse) for X in self.tensors)


class CPTensor(Tensor):
    """Implements CP tensors.

    The cores are stored as list of shape ``(shape[i], rank)``."""

    shape: Tuple[int, ...]
    rank: int
    cores: ArrayList

    def __init__(self, cores: ArrayList) -> None:
        self.cores = cores
        self.rank = cores[0].shape[1]
        self.shape = tuple(C.shape[0] for C in cores)

    def size(self) -> int:
        return sum(C.size for C in self.cores)

    @property
    def T(self) -> CPTensor:
        new_cores = self.cores[::-1]
        return self.__class__(new_cores)

    def to_numpy(self) -> npt.NDArray:
        dense_tensor = self.cores[0]  # shape (n0,r)
        for C in self.cores[1:]:
            dense_tensor = np.einsum(
                "...j,ij->...ij", dense_tensor, C
            )  # shape (n0,...,nk,r)
        dense_tensor = np.sum(dense_tensor, axis=-1)
        return dense_tensor

    @classmethod
    def random(
        cls, shape: Tuple[int, ...], rank: int, seed: Optional[int] = None
    ) -> CPTensor:
        d = len(shape)
        seq = SeedSequence(seed)
        seeds = seq.generate_state(d)
        cores = []
        for i in range(d):
            core = random_normal(shape=(shape[i], rank), seed=seeds[i])
            core /= np.sqrt(shape[i])
            cores.append(core)

        return cls(cores)

    def __getitem__(self, index: int) -> npt.NDArray:
        return self.cores[index]

    def __setitem__(self, index: int, data: npt.NDArray) -> None:
        self.cores[index] = data

    def gather(self, idx: Tuple[npt.NDArray[np.int64], ...]) -> npt.NDArray:
        """Obtain the values of the tensor at the given indices."""

        res = 1
        for C, id in zip(self.cores, idx):
            res *= C[id]
        return np.sum(res, axis=1)

    def __repr__(self) -> str:
        return (
            f"<CP tensor of shape {self.shape} and rank {self.rank} "
            f"at {hex(id(self))}>"
        )

    def __mul__(self, other: float) -> CPTensor:
        new_cores = [c for c in self.cores]
        new_cores[0] = new_cores[0] * other
        return self.__class__(new_cores)


class TuckerTensor(Tensor):
    """Implements Tucker tensors.

    This consists of a core tensor of shape ``(s1, ..., sd)`` and ``d``
    factor matrices of shape ``(si, ni)``, where ``(n1, ..., nd)`` is the
    overall shape of the tensor."""

    def __init__(self, factors: ArrayList, core: npt.NDArray) -> None:
        self.core = core
        self.factors = factors

        self.shape = tuple(U.shape[1] for U in factors)
        self.rank = tuple(U.shape[0] for U in factors)

    @property
    def T(self) -> TuckerTensor:
        new_factors = self.factors[::-1]
        permutation = tuple(range(len(self.shape))[::-1])
        new_core = np.transpose(self.core, permutation)
        return self.__class__(new_factors, new_core)

    @property
    def size(self) -> int:
        return self.core.size + sum(U.size for U in self.factors)

    def to_numpy(self) -> npt.NDArray[np.float64]:
        core_contracted = self.core
        for i, U in enumerate(self.factors):
            left_dim = np.prod(self.shape[:i], dtype=np.int64)
            right_dim = np.prod(self.rank[i + 1 :], dtype=np.int64)
            core_mat = core_contracted.reshape(
                left_dim, self.rank[i], right_dim
            )
            core_contracted = np.einsum("ijk,jl->ilk", core_mat, U)
        return core_contracted.reshape(self.shape)

    def __mul__(self, other: float) -> TuckerTensor:
        new_core = self.core * other
        return self.__class__(self.factors, new_core)

    def __repr__(self) -> str:
        return (
            f"<Tucker tensor of shape {self.shape} and rank {self.rank} "
            f"at {hex(id(self))}>"
        )

    @classmethod
    def random(
        cls,
        shape: Tuple[int, ...],
        rank: Union[int, Tuple[int, ...]],
        seed: Optional[int] = None,
    ) -> TuckerTensor:
        d = len(shape)
        try:
            rank_tuple = tuple(rank)  # type: ignore
        except TypeError:
            rank_tuple = (rank,) * d  # type: ignore
        rank_tuple = tuple(min(r1, r2) for r1, r2 in zip(rank_tuple, shape))

        seq = SeedSequence(seed)
        core_seed = seq.generate_state(1)[0]
        core = random_normal(shape=rank_tuple, seed=core_seed)
        factors = []
        seeds = seq.generate_state(d)
        for r, n, seed in zip(rank_tuple, shape, seeds):
            U = random_normal(shape=(r, n), seed=seed)
            U = np.linalg.qr(U.T)[0].T
            factors.append(U)

        return cls(factors, core)



import time
from functools import reduce
from os.path import isfile
from typing import Callable, Dict, Optional, Any

import pandas as pd
from tt_sketch.sketch import stream_sketch, orthogonal_sketch, hmt_sketch
from tt_sketch.tensor import Tensor, TensorTrain
from tt_sketch.tt_svd import tt_svd


class Experiment:
    data: pd.DataFrame
    filename: str
    autosave: bool

    def __init__(
        self,
        filename: str,
        data: Optional[pd.DataFrame] = None,
        force_overwrite: bool = False,
    ):
        self.filename = filename
        if data is None:
            if isfile(filename) and not force_overwrite:
                data = pd.read_csv(filename)
            else:
                data = pd.DataFrame(columns=["name", "error", "time_taken"])

        self.data = data

    @classmethod
    def load(cls, filename):
        data = pd.read_csv(filename)
        return cls(filename, data)

    def save(self):
        self.data.to_csv(self.filename)

    def check_experiment_done(self, conds):
        masks = []
        for column, value in conds.items():
            try:
                masks.append(self.data[column] == value)
            except KeyError:
                return False
        mask = reduce(lambda x, y: x & y, masks)
        return mask.sum() > 0

    def do_experiment(
        self,
        input,
        name: str,
        experiment_func: Callable[..., Dict[str, Any]],
        **kwargs,
    ):
        """Do an experiment, and store results+metadata.

        ``input`` should be something tensor-like that ``experiment_func``
        understands.

        ``experiment_func`` has signature ``(input, **kwargs) -> Dict[str, Any]``.
        """

        row = dict()
        for key, value in kwargs.items():
            try:  # Check if values is numeric
                float(value)
            except (ValueError, TypeError):
                if not isinstance(value, str):
                    try:
                        value = value.__name__
                    except AttributeError:
                        value = str(value)
            row[key] = value
        row["name"] = name
        if self.check_experiment_done(row):
            return

        # Do the experiment
        result = experiment_func(input, **kwargs)
        for key, value in result.items():
            row[key] = value
        # row["error"] = result

        row_df = pd.DataFrame([row.values()], columns=row.keys())
        self.data = pd.concat([self.data, row_df], ignore_index=True)

        self.save()


def experiment_stream_sketch(
    input_tensor: Tensor,
    left_rank=None,
    right_rank=None,
    left_drm_type=None,
    right_drm_type=None,
    error_func=None,
    recompression_rank=None,
    **kwargs,
) -> Dict[str, Any]:
    start_time = time.perf_counter()
    tt_sketched = stream_sketch(
        input_tensor,
        left_rank=left_rank,
        right_rank=right_rank,
        left_drm_type=left_drm_type,
        right_drm_type=right_drm_type,
    )
    if recompression_rank is not None:
        tt_sketched = tt_sketched.to_tt().round(max_rank=recompression_rank)
        assert max(tt_sketched.rank) <= recompression_rank
    time_taken = time.perf_counter() - start_time

    if error_func is not None:
        error = error_func(tt_sketched, input_tensor)
    else:
        error = tt_sketched.error(input_tensor, relative=True)
    return {"error": error, "time_taken": time_taken}


def experiment_orthogonal_sketch(
    input_tensor: Tensor,
    left_rank=None,
    right_rank=None,
    left_drm_type=None,
    right_drm_type=None,
    error_func=None,
    recompression_rank=None,
    **kwargs,
) -> Dict[str, Any]:
    start_time = time.perf_counter()
    tt_sketched = orthogonal_sketch(
        input_tensor,
        left_rank=left_rank,
        right_rank=right_rank,
        left_drm_type=left_drm_type,
        right_drm_type=right_drm_type,
    )
    if recompression_rank is not None:
        tt_sketched = tt_sketched.round(max_rank=recompression_rank)
    time_taken = time.perf_counter() - start_time

    if error_func is not None:
        error = error_func(tt_sketched, input_tensor)
    else:
        error = tt_sketched.error(input_tensor, relative=True)
    return {"error": error, "time_taken": time_taken}


def experiment_hmt_sketch(
    input_tensor: Tensor,
    rank=None,
    drm_type=None,
    error_func=None,
    recompression_rank=None,
    **kwargs,
):
    start_time = time.perf_counter()
    tt_sketched = hmt_sketch(input_tensor, rank=rank, drm_type=drm_type)
    if recompression_rank is not None:
        tt_sketched = tt_sketched.round(max_rank=recompression_rank)
    time_taken = time.perf_counter() - start_time
    if error_func is not None:
        error = error_func(tt_sketched, input_tensor)
    else:
        error = tt_sketched.error(input_tensor, relative=True)
    return {"error": error, "time_taken": time_taken}


def experiment_tt_svd(
    input_tensor: Tensor, rank=None, error_func=None, **kwargs
) -> Dict[str, Any]:
    start_time = time.perf_counter()
    tt = tt_svd(input_tensor, rank=rank)
    time_taken = time.perf_counter() - start_time
    if error_func is not None:
        error = error_func(tt, input_tensor)
    else:
        error = tt.error(input_tensor, relative=True)
    return {"error": error, "time_taken": time_taken}

def experiment_tt_round(
    input_tensor: TensorTrain, rank=None, error_func=None, **kwargs
) -> Dict[str, Any]:
    start_time = time.perf_counter()
    tt = input_tensor.round(max_rank=rank)
    time_taken = time.perf_counter() - start_time
    if error_func is not None:
        error = error_func(tt, input_tensor)
    else:
        error = tt.error(input_tensor, relative=True)
    return {"error": error, "time_taken": time_taken}



# %%
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


csv_filename = "results/cookie_2x2.csv"

df = pd.read_csv(csv_filename, index_col=0)
df.drop(columns=df.columns[0:4], inplace=True, axis=1)

for c in ["final_round_ranks"]:
    df[c] = df[c].apply(eval).apply(np.array)


def process_sl(sl):
    return np.array(
        eval(
            ",".join([s.replace("\n", "") for s in sl])
            .replace(",,", ",")
            .replace(",,", ",")
        )
    )


for c in ["true_error", "precond_error", "final_round_time"]:
    df[c] = df[c].str.split(" ").apply(process_sl)

initial_columns = [
    "rounding_method",
    "rounding_method_final",
    "max_rank",
    "run",
    "gmres_time",
]
initial_df = df[initial_columns]
final_columns = [
    "final_round_ranks",
    "true_error",
    "precond_error",
    "final_round_time",
]
final_df = df[final_columns]

row_dfs = []
for row in final_df.itertuples():
    row_idx = row[0]
    row_df = pd.DataFrame(np.stack(row[1:], axis=1), columns=final_columns)
    row_df["row_idx"] = row_idx
    row_dfs.append(row_df)

row_dfs = pd.concat(row_dfs)
initial_df.index.name = "row_idx"
plot_df = initial_df.merge(row_dfs, on="row_idx")
plot_df["total_time"] = plot_df["gmres_time"] + plot_df["final_round_time"]
# %%
for rounding_method in ["sketch", "pairwise"]:
    plot_df_rm = plot_df[plot_df["rounding_method"] == rounding_method]
    plot_df.groupby(["rounding_method", "max_rank"])["true_error"].min()
    plot_df_rm = plot_df_rm.sort_values("total_time")
    cummin = plot_df_rm["true_error"].cummin()
    plt.plot(plot_df_rm["total_time"], cummin, label=rounding_method)

plt.yscale("log")
plt.xscale("log")
plt.title("Relative error vs. time for rounding methods used in body of TT-GMRES")
plt.legend()
plt.ylabel("Relative error")
plt.xlabel("Total time")



"""Implements the TT-GMRES algorithm for solving linear systems in the 
TT-format, as described in Dolgov, arXiv:1206.5512, but with rounding step
optionally replaced by sketching."""
from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from copy import deepcopy
import logging
from math import ceil
from time import perf_counter
from typing import Dict, List, Literal, Optional, Tuple, Union, Any

import numpy as np
import numpy.typing as npt
import scipy.linalg

from tt_sketch.sketch import orthogonal_sketch, stream_sketch
from tt_sketch.tensor import Tensor, TensorSum, TensorTrain
from tt_sketch.utils import (
    ArrayList,
    TTRank,
    dematricize,
    matricize,
    process_tt_rank,
)


class TTLinearMap(ABC):
    """Abstract class for linear maps in the TT-format."""

    in_shape: Tuple[int, ...]
    out_shape: Tuple[int, ...]

    @abstractmethod
    def __call__(self, other: TensorTrain) -> TensorTrain:
        ...


class MPO(Tensor, TTLinearMap):
    """MPO with order 4 tensor cores of shape
    ``(rank[mu-1],in_shape[mu],out_shape[mu],rank[mu])``

    Used as linear map in the tensor train format"""

    in_shape: Tuple[int, ...]
    out_shape: Tuple[int, ...]
    rank: Tuple[int, ...]
    shape: Tuple[int, ...]
    cores: ArrayList

    def __init__(self, cores: ArrayList) -> None:
        self.cores = cores
        self.in_shape = tuple(C.shape[1] for C in cores)
        self.out_shape = tuple(C.shape[2] for C in cores)
        self.rank = tuple(C.shape[0] for C in cores[1:])
        self.shape = tuple(
            s1 * s2 for s1, s2 in zip(self.in_shape, self.out_shape)
        )

    @property
    def size(self) -> int:
        return sum(C.size for C in self.cores)

    @property
    def T(self) -> MPO:
        """Transposition here is that of a linear map, this is different from
        other tensors."""
        new_cores = [C.transpose((0, 2, 1, 3)) for C in self.cores]
        return self.__class__(new_cores)

    def to_tt(self) -> TensorTrain:
        new_cores = [
            C.reshape(C.shape[0], C.shape[1] * C.shape[2], C.shape[3])
            for C in self.cores
        ]
        return TensorTrain(new_cores)

    def to_numpy(mpo) -> npt.NDArray:
        """Contract to dense array of shape
        ``(in_shape[0], out_shape[0], ..., in_shape[d-1], outs_shape[d-1])``"""

        res = mpo.cores[0]
        res = res.reshape(res.shape[1:])
        for C in mpo.cores[1:]:
            res = np.einsum("...i,ijkl->...jkl", res, C)
        res = res.reshape(res.shape[:-1])
        return res

    def __call__(self, other: TensorTrain) -> TensorTrain:
        new_cores = []
        for M, C in zip(self.cores, other.cores):
            MC = np.einsum("ijkl,ajb->iaklb", M, C)
            MC = MC.reshape(
                MC.shape[0] * MC.shape[1],
                MC.shape[2],
                MC.shape[3] * MC.shape[4],
            )
            new_cores.append(MC)
        return TensorTrain(new_cores)

    @classmethod
    def random(
        cls,
        rank: TTRank,
        in_shape: Tuple[int, ...],
        out_shape: Tuple[int, ...],
    ) -> MPO:
        prod_shape = tuple(s1 * s2 for s1, s2 in zip(in_shape, out_shape))
        rank = process_tt_rank(rank, prod_shape, trim=True)
        cores = []
        for r1, s1, s2, r2 in zip(
            (1,) + rank, in_shape, out_shape, rank + (1,)
        ):
            C = np.random.normal(size=(r1, s1, s2, r2))
            C += C.transpose(0, 2, 1, 3).reshape(C.shape)  # symmetrize
            C = C * np.sqrt(s1 * s2) / np.linalg.norm(C)
            cores.append(C)
        return cls(cores)

    @classmethod
    def eye(cls, shape) -> MPO:
        cores = []
        for s in shape:
            C = np.eye(s, s)
            C = C.reshape(1, C.shape[0], C.shape[1], 1)
            cores.append(C)
        return cls(cores)

    def __mul__(self, other: float) -> MPO:
        new_cores = self.cores
        new_cores[0] = new_cores[0] * other
        return self.__class__(new_cores)


class TTPrecond(TTLinearMap):
    """TTLinearMap that acts by multiplying by the inverse of a matrix on a
    specified mode.

    The inverse is computed from the precomputed QR factorization of the matrix.
    """

    def __init__(self, A, shape, mode=0):
        self.A = A
        self.Q, self.R = np.linalg.qr(A)
        self.mode = mode
        self.in_shape = shape
        self.out_shape = shape

    def backward_call(self, other: TensorTrain) -> TensorTrain:
        new_cores = deepcopy(other.cores)
        C = new_cores[self.mode]
        C_mat = matricize(C, mode=1, mat_shape=True)
        sol = scipy.linalg.solve_triangular(self.R, (self.Q.T) @ C_mat)
        new_cores[self.mode] = dematricize(sol, mode=1, shape=C.shape)
        return TensorTrain(new_cores)

    def forward_call(self, other: TensorTrain) -> TensorTrain:
        new_cores = deepcopy(other.cores)
        C = new_cores[self.mode]
        C_mat = matricize(C, mode=1, mat_shape=True)
        new_cores[self.mode] = dematricize(
            self.A @ C_mat, mode=1, shape=C.shape
        )
        return TensorTrain(new_cores)

    __call__ = backward_call


class TTLinearMapSum:
    """Linear map that eats a TT and returns a sum of TTs.

    This is essentially just a container for a list of ``TTLinearMap`` objects.
    """

    in_shape: Tuple[int, ...]
    out_shape: Tuple[int, ...]
    linear_maps: List[TTLinearMap]

    def __init__(self, linear_maps: List[TTLinearMap]) -> None:
        self.linear_maps = linear_maps
        if len(linear_maps) == 0:
            raise ValueError("linear_maps cannot be empty")
        self.in_shape = linear_maps[0].in_shape
        self.out_shape = linear_maps[0].out_shape
        for linear_map in linear_maps[1:]:
            if linear_map.in_shape != self.in_shape:
                raise ValueError("in_shape mismatch")
            if linear_map.out_shape != self.out_shape:
                raise ValueError("out_shape mismatch")

    def __call__(
        self, input_tensor: Union[TensorTrain, TensorSum[TensorTrain]]
    ) -> TensorSum[TensorTrain]:
        if isinstance(input_tensor, TensorTrain):
            tensor_list = [input_tensor]
        else:
            tensor_list = input_tensor.tensors

        output_list = []
        for linear_map in self.linear_maps:
            for tensor in tensor_list:
                output_list.append(linear_map(tensor))

        return TensorSum(output_list)


# def tt_weighted_sum_sketched(
#     x0: TensorTrain,
#     coeffs: npt.NDArray,
#     tt_list: List[TensorTrain],
#     tolerance: float,
#     max_rank: Tuple[int, ...],
#     round: bool = False,
# ):
#     """Sketched weighted sum of tensor trains."""
#     x_sum = TensorSum([x0])
#     for coeff, tt in zip(coeffs, tt_list):
#         x_sum += coeff * tt
#     x = round_tt_sum(x_sum, tolerance, max_rank, False)
#     return x


# def tt_weighted_sum_exact(
#     x0: TensorTrain,
#     coeffs: npt.NDArray,
#     tt_list: List[TensorTrain],
#     tolerance: float,
#     max_rank: Tuple[int, ...],
# ):
#     """Weighted sum of tensor trains rounded to ``max_rank``"""
#     x = x0
#     for coeff, tt in zip(coeffs, tt_list):
#         x = x.add(coeff * tt)
#     x = x.round(tolerance, max_rank)

#     return x


# def tt_weighted_sum(
#     x0: TensorTrain,
#     coeffs: npt.NDArray,
#     tt_list: List[TensorTrain],
#     tolerance: float,
#     max_rank: Tuple[int, ...],
#     exact: bool = False,
#     oversample: int = 5,
# ):
#     x_sum = TensorSum([x0])
#     for coeff, tt in zip(coeffs, tt_list):
#         x_sum += coeff * tt
#     x = round_tt_sum(x_sum, max_rank, tolerance, exact, oversample)
#     return x


ROUNDING_MODE = Literal["exact", "pairwise", "sketch", "orth_sketch", None]


def round_tt_sum(
    tt_sum: TensorSum[TensorTrain],
    max_rank: TTRank,
    eps: Optional[float] = None,
    method: ROUNDING_MODE = "sketch",
    oversample_factor: float = 2,
) -> TensorTrain:
    """Round a sum of tensor trains to a given rank.

    method can be one of:
    - "exact": Add all TTs to one big TT and round it using TT-SVD
    - "pairwise": Add each TT to the next one and round them separately
    - "sketch": Use streaming sketch for rounding
    - "orth_sketch": Use orthogonal sketch for rounding.
    - ``None``: Do not round (for debugging purposes mostly).
    """
    if method == "exact":
        summands = tt_sum.tensors
        tt = summands[0]
        for summand in summands[1:]:
            tt = tt.add(summand)
        return tt.round(eps, max_rank)
    elif method == "pairwise":
        tt = tt_sum.tensors[0]
        for t in tt_sum.tensors[1:]:
            tt = tt.add(t).round(eps=eps, max_rank=max_rank)
    elif method == "sketch":
        left_rank = process_tt_rank(max_rank, tt_sum.shape, trim=True)
        right_rank = tuple(ceil(r * oversample_factor) for r in left_rank)

        tt = stream_sketch(
            tt_sum, left_rank=left_rank, right_rank=right_rank
        ).to_tt()
    elif method == "orth_sketch":
        left_rank = process_tt_rank(max_rank, tt_sum.shape, trim=True)
        right_rank = tuple(ceil(r * oversample_factor) for r in left_rank)

        tt = orthogonal_sketch(
            tt_sum, left_rank=left_rank, right_rank=right_rank
        )
    elif method is None:
        return tt_sum  # type: ignore
    else:
        raise ValueError(f"Unknown rounding method: {method}")

    return tt


def tt_sum_gmres(
    A: TTLinearMapSum,
    b: TensorTrain,
    max_rank: TTRank,
    precond: Optional[TTPrecond] = None,
    final_round_rank: Optional[TTRank] = None,
    x0: Optional[TensorTrain] = None,
    tolerance: float = 1e-6,
    maxiter: int = 100,
    symmetric: bool = False,
    rounding_method: ROUNDING_MODE = "pairwise",
    rounding_method_final: Optional[ROUNDING_MODE] = None,
    save_basis: bool = False,
    verbose: bool = False,
) -> Tuple[TensorTrain, Dict[str, List]]:
    """
    GMRES solver for TTLinearMapSum.

    The TTLinearMapSum takes as input a TT and returns a sum of TTs. This means
    additional rounding steps are required in comparison to a version of
    TT-GMRES where the output of the linear map is a TT.
    """
    if final_round_rank is None:
        final_round_rank = max_rank
    if rounding_method_final is None:
        rounding_method_final = rounding_method
    if A.out_shape != b.shape:
        raise ValueError("Output shape of linear map doesn't match RHS")
    if x0 is not None and x0.shape != A.in_shape:
        raise ValueError("Input shape of liner map doesn't match initial value")
    if A.out_shape != A.in_shape:
        raise ValueError("TT-GMRES only works for automorphisms")

    max_rank = process_tt_rank(max_rank, A.in_shape, trim=True)
    if x0 is None:
        # TODO: check whether init with zero or random init is better
        # x0 = TensorTrain.random(shape=A.in_shape, rank=max_rank)
        x0 = TensorTrain.zero(shape=A.in_shape, rank=1)

    def apply_A_pr(x: TensorTrain) -> TensorSum[TensorTrain]:
        res = A(x)
        if precond is not None:
            res = TensorSum([precond(r) for r in res.tensors])
        return res

    b_pr = precond(b) if precond is not None else b

    b_norm = b.norm()
    initial_time = perf_counter()
    residual = b_pr - apply_A_pr(x0)
    residual_rounded = round_tt_sum(
        residual, max_rank=max_rank, method=rounding_method  # type: ignore
    )
    residual_norm = residual_rounded.norm()
    beta = residual_norm
    nu_list: List[TensorTrain] = [residual_rounded / beta]
    H_matrix = np.zeros((maxiter + 1, maxiter))

    history: Dict[str, Any] = defaultdict(list)
    history["w_norm"].append(nu_list[-1].norm())
    history["rank"].append(residual_rounded.rank)
    history["residual_norm"].append(residual_norm / b_norm)
    history["step_time"].append(perf_counter() - initial_time)

    for j in range(maxiter):
        current_time = perf_counter()
        delta = tolerance / (residual_norm / beta)
        if verbose:
            logging.info(
                f"Iteration {j + 1}/{maxiter}, residual norm: {residual_norm / b_norm:.4e}"
            )
        w_sum = apply_A_pr(nu_list[-1])
        w_rounded = round_tt_sum(
            w_sum, eps=delta, max_rank=max_rank, method=rounding_method
        )

        min_j = max(0, j - 2) if symmetric else 0
        for i in range(min_j, j + 1):
            H_matrix[i, j] = w_rounded.dot(nu_list[i])

        # Do Gram-Schmidt orthogonalization
        w_sum = (
            w_rounded
            - TensorSum(nu_list[min_j : j + 1]) * H_matrix[min_j : j + 1, j]
        )
        w_rounded = round_tt_sum(
            w_sum, eps=delta, max_rank=max_rank, method=rounding_method
        )
        H_matrix[j + 1, j] = w_rounded.norm()
        nu_list.append(w_rounded / H_matrix[j + 1, j])
        history["step_time"].append(perf_counter() - current_time)

        # Compute residual norm
        H_red = H_matrix[: j + 2, : j + 1]
        e1 = np.zeros(j + 2)
        e1[0] = beta
        y, (residual_norm,), _, _ = np.linalg.lstsq(H_red, e1, rcond=None)
        history["step_time_with_res_norm"].append(perf_counter() - current_time)
        history["residual_norm"].append(np.sqrt(residual_norm) / b_norm)
        history["rank"].append(w_rounded.rank)
        history["w_norm"].append(H_matrix[j + 1, j])
        history["delta"].append(delta)

        if residual_norm / b_norm < tolerance:
            break

    # Compute final result and round
    y = y[: j + 1]
    nu_list = nu_list[: j + 1]
    current_time = perf_counter()
    result = x0 + TensorSum(nu_list) * y
    result_rounded = round_tt_sum(
        result,
        eps=None,
        max_rank=final_round_rank,
        method=rounding_method_final,
    )

    history["final_round_time"] = perf_counter() - current_time
    history["total_time"] = perf_counter() - initial_time
    if save_basis:
        history["H_matrix"] = H_matrix
        history["nu_list"] = nu_list
        history["y"] = y
    return result_rounded, history



from __future__ import annotations
from copy import deepcopy

from typing import Optional, Tuple

import numpy as np

from tt_sketch.utils import ArrayList


class SketchContainer:
    """
    Container class for the Psi_cores and Omega_mats often used internally.
    """

    shape: Tuple[int, ...]
    left_rank: Tuple[int, ...]
    right_rank: Tuple[int, ...]
    Psi_cores: ArrayList
    Omega_mats: ArrayList

    def __init__(
        self,
        Psi_cores: ArrayList,
        Omega_mats: ArrayList,
        shape: Optional[Tuple[int, ...]] = None,
        left_rank: Optional[Tuple[int, ...]] = None,
        right_rank: Optional[Tuple[int, ...]] = None,
    ) -> None:
        self.Psi_cores = Psi_cores
        self.Omega_mats = Omega_mats

        # Infer shapes and ranks from Psi_cores
        if shape is None:
            shape = tuple(Psi.shape[1] for Psi in Psi_cores)
        if left_rank is None:
            left_rank = tuple(Psi.shape[0] for Psi in Psi_cores[1:])
        if right_rank is None:
            right_rank = tuple(Psi.shape[2] for Psi in Psi_cores[:-1])
        self.shape = shape
        self.left_rank = left_rank
        self.right_rank = right_rank

    @classmethod
    def zero(
        cls,
        shape: Tuple[int, ...],
        left_rank: Tuple[int, ...],
        right_rank: Tuple[int, ...],
    ) -> SketchContainer:
        Psi_cores = []
        for r1, n, r2 in zip((1,) + left_rank, shape, right_rank + (1,)):
            Psi_cores.append(np.zeros((r1, n, r2)))

        Omega_mats = []
        for r1, r2 in zip(left_rank, right_rank):
            Omega_mats.append(np.zeros((r1, r2)))

        return cls(Psi_cores, Omega_mats, shape, left_rank, right_rank)

    def __add__(self, other: SketchContainer) -> SketchContainer:
        Psi_cores_new = [
            Psi1 + Psi2 for Psi1, Psi2 in zip(self.Psi_cores, other.Psi_cores)
        ]
        Omega_mats_new = [
            Omega1 + Omega2
            for Omega1, Omega2 in zip(self.Omega_mats, other.Omega_mats)
        ]
        return self.__class__(Psi_cores_new, Omega_mats_new)

    @property
    def T(self) -> SketchContainer:
        Psi_cores_new = [Psi.transpose(2, 1, 0) for Psi in self.Psi_cores[::-1]]
        Omega_mats_new = [Omega.T for Omega in self.Omega_mats[::-1]]
        return self.__class__(Psi_cores_new, Omega_mats_new)

    def __mul__(self, other: float) -> SketchContainer:
        new_Psi_cores = deepcopy(self.Psi_cores)
        new_Psi_cores[0] *= other
        return self.__class__(new_Psi_cores, self.Omega_mats)
